==> spread-5.3.0.5-py3-none-any.whl: spread_core/bam/handling/items/triggers.py <==
import datetime
import json

import spread_core.mqtt as mqtt
from spread_core.tools import utils
from spread_core.tools.settings import config, CALENDAR
from .. import const, errors

WORKDAYS_TYPE = 'workdays'
WEEKENDS_TYPE = 'weekends'

SUNDAY = 'Su'
MONDAY = 'Mo'
TUESDAY = 'Tu'
WEDNESDAY = 'We'
THURSDAY = 'Th'
FRIDAY = 'Fr'
SATURDAY = 'Sa'

WORKDAYS = [MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY]
WEEKENDS = [SATURDAY, SUNDAY]
WEEK = WORKDAYS + WEEKENDS


def generate_holidays():
    today = datetime.datetime.now().date()
    result = []

    calendar_path = config[CALENDAR]
    if not calendar_path:
        raise BaseException('Calendar path is not set!')
    with open(calendar_path, 'r') as file:
        data = file.read()
    data = json.loads(data)
    for date_str in data:
        _date = datetime.date.fromisoformat(date_str)
        if _date >= today:
            result.append(_date)

    result.sort()

    return result


class Trigger:
    TYPE = None

    def __init__(self, *args, **kwargs):
        self._id = int(kwargs['id'])

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.TYPE == other.TYPE

    def __hash__(self):
        return self.id

    @property
    def id(self):
        return self._id

    def serialize(self) -> dict:
        return {const.TYPE: self.TYPE, const.ID: self.id}


class _AnyValue:
    VALUE = '#any'

    def __eq__(self, other):
        return True

    def serialize(self):
        return self.VALUE


class _RangeValue:
    VALUE = '#range('

    def __init__(self, ranges: str):
        self.ranges = []
        try:
            for i in ranges.split(','):
                i = i.strip()
                bounds = i.split('-')
                self.ranges.append(range(int(bounds[0]), int(bounds[1])))
        except BaseException as ex:
            raise errors.ParseError(f'Ошибка парсинга {self.__class__.__name__} для {ranges}: {ex}')

    def __eq__(self, other):
        for _range in self.ranges:
            if other in _range:
                return True

        return False

    def serialize(self) -> str:
        ranges = [f'{r.start}-{r.stop}' for r in self.ranges]
        return f'{self.VALUE}{",".join(ranges)})'


class _ListValue:
    VALUE = '#list('

    def __init__(self, items: str):
        try:
            self.items = []
            for i in items.split(','):
                i = i.strip()
                if i.isdigit():
                    i = int(i)
                self.items.append(i)
        except BaseException as ex:
            raise errors.ParseError(f'Ошибка парсинга {self.__class__.__name__} для {items}: {ex}')

    def serialize(self) -> str:
        return f'{self.VALUE}{",".join(str(i) for i in self.items)})'

    def __eq__(self, other):
        return other in self.items


class BrokerTrigger(Trigger):
    TYPE = 'broker'

    def __init__(self, *args, **kwargs):
        super(BrokerTrigger, self).__init__(*args, **kwargs)
        self._topic = mqtt.of(kwargs[const.TOPIC])
        _value = kwargs[const.VALUE]
        if isinstance(_value, str) and _value == _AnyValue.VALUE:
            self._value = _AnyValue()
        elif isinstance(_value, str) and _value.endswith(')') and _value.startswith(_RangeValue.VALUE):
            self._value = _RangeValue(_value[len(_RangeValue.VALUE):-1])
        elif isinstance(_value, str) and _value.endswith(')') and _value.startswith(_ListValue.VALUE):
            self._value = _ListValue(_value[len(_ListValue.VALUE):-1])
        else:
            self._value = _value

    def serialize(self) -> dict:
        res = super(BrokerTrigger, self).serialize()
        res[const.TOPIC] = str(self._topic)
        if isinstance(self._value, (_AnyValue, _RangeValue, _ListValue)):
            res[const.VALUE] = self._value.serialize()
        else:
            res[const.VALUE] = self._value

        return res

    def __str__(self):
        return f'{self._topic} as {self._value}'

    def check(self, topic: str, variable: mqtt.spread.variable.Variable):
        if str(topic) == str(self._topic):
            if variable.value == self._value:
                return True

        return False

    @property
    def topic(self) -> mqtt.spread.topic.SpreadTopic: return self._topic

    @property
    def value(self): return self._value


class DateTrigger(Trigger):
    def __init__(self, *args, **kwargs):
        super(DateTrigger, self).__init__(*args, **kwargs)
        self.time = datetime.time.fromisoformat(kwargs[const.TIME])

    @property
    def is_ready(self):
        return False

    @property
    def has_next(self):
        return False

    @property
    def time_left(self):
        return 1

    def serialize(self) -> dict:
        res = super(DateTrigger, self).serialize()
        res[const.TIME] = str(self.time)
        return res

    def __str__(self):
        return f'<{self.id}>{self.days if isinstance(self, WeekDays) else self.TYPE} at {self.time}'

    def __eq__(self, other):
        return super(DateTrigger, self).__eq__(other) and self.time == other.time

    def __cmp__(self, other):
        return self.time_left - other.time_left

    def __lt__(self, other):
        return self.time_left < other.time_left

    def __hash__(self):
        return self.id


class WeekDays(DateTrigger):
    TYPE = 'weekdays'

    def __init__(self, *args, **kwargs):
        super(WeekDays, self).__init__(*args, **kwargs)
        days = kwargs[const.DAYS]
        self.days = []
        if len(days) == 0:
            raise errors.HandlingError(f'{const.DAYS} expected 1..7 items')
        for day in days:
            if day in WEEK:
                self.days.append(day)
            else:
                raise errors.HandlingError(f'{day} not include in standard values {WEEK}')

    def serialize(self) -> dict:
        res = super(WeekDays, self).serialize()
        res[const.DAYS] = self.days
        return res

    @property
    def has_next(self):
        return True

    @property
    def time_left(self):
        now = datetime.datetime.now()
        days = []
        for day in self.days:
            index = WEEK.index(day)
            if index == now.weekday() and self.time > now.time():
                delta = 0
            elif index > now.weekday() or (index == now.weekday() and self.time > now.time()):
                delta = index - now.weekday()
            else:
                delta = 7 - now.weekday() + index
            _date = (now + datetime.timedelta(days=delta)).date()
            t = datetime.datetime.fromisoformat(str(_date) + ' ' + str(self.time))
            days.append(t)

        days.sort()

        target_time = days[0].timestamp()
        return int(target_time - now.timestamp())


class Holiday(DateTrigger):
    TYPE = 'holiday'
    _days = None

    @staticmethod
    def days():
        if Holiday._days is None:
            Holiday._days = generate_holidays()
        return Holiday._days

    @property
    def has_next(self):
        return True

    @property
    def time_left(self):
        now = datetime.datetime.now()

        days = self.days().copy()

        while len(days) > 0:
            if days[0] < now.date() or (days[0] == now.date() and now.time() > self.time):
                days.pop(0)
            else:
                break

        if len(days) > 0:
            _date = days[0]
        else:
            raise errors.HandlingError('calendar is empty!')

        target_time = datetime.datetime.fromisoformat(str(_date) + ' ' + str(self.time))
        return int(target_time.timestamp() - now.timestamp())


class WorkDays(DateTrigger):
    TYPE = 'workday'

    @property
    def has_next(self):
        return True

    @property
    def time_left(self):
        now = datetime.datetime.now()

        _date = now.date()

        if _date not in Holiday.days():
            if self.time <= now.time():
                _date += datetime.timedelta(days=1)

        while _date in Holiday.days():
            _date += datetime.timedelta(days=1)

        target_time = datetime.datetime.fromisoformat(str(_date) + ' ' + str(self.time))
        return int((target_time - now).total_seconds())


def date_trigger_of_data(_type: str, data: dict) -> DateTrigger:
    for cl in DateTrigger.__subclasses__():
        if issubclass(cl, DateTrigger):
            if cl.TYPE == _type:
                return cl(**data)


def of(data: dict) -> Trigger:
    def separator(subclass): return issubclass(subclass, Trigger) and subclass.TYPE == data[const.TYPE]
    trigger_class = utils.get_subclass(Trigger, separator=separator)
    return trigger_class(**data)

==> spread-5.9.0.59-py3-none-any.whl: spread_core/bam/handling/items/triggers.py <==
import datetime
import json

import spread_core.mqtt as mqtt
from spread_core.tools import utils
from spread_core.tools.settings import config, CALENDAR
from .. import const, errors

WORKDAYS_TYPE = 'workdays'
WEEKENDS_TYPE = 'weekends'

SUNDAY = 'Su'
MONDAY = 'Mo'
TUESDAY = 'Tu'
WEDNESDAY = 'We'
THURSDAY = 'Th'
FRIDAY = 'Fr'
SATURDAY = 'Sa'

WORKDAYS = [MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY]
WEEKENDS = [SATURDAY, SUNDAY]
WEEK = WORKDAYS + WEEKENDS


def generate_holidays():
    today = datetime.datetime.now().date()
    result = []

    calendar_path = config[CALENDAR]
    if not calendar_path:
        raise BaseException('Calendar path is not set!')
    with open(calendar_path, 'r') as file:
        data = file.read()
    data = json.loads(data)
    for date_str in data:
        _date = datetime.date.fromisoformat(date_str)
        if _date >= today:
            result.append(_date)

    result.sort()

    return result


class Trigger:
    TYPE = None

    def __init__(self, *args, **kwargs):
        self._id = int(kwargs['id'])

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.TYPE == other.TYPE

    def __hash__(self):
        return self.id

    @property
    def id(self):
        return self._id

    def serialize(self) -> dict:
        return {const.TYPE: self.TYPE, const.ID: self.id}


class _AnyValue:
    VALUE = '#any'

    def __eq__(self, other):
        return True

    def serialize(self):
        return self.VALUE


class _RangeValue:
    VALUE = '#range('

    def __init__(self, ranges: str):
        self.ranges = []
        try:
            for i in ranges.split(','):
                i = i.strip()
                bounds = i.split('-')
                self.ranges.append(range(int(bounds[0]), int(bounds[1])))
        except BaseException as ex:
            raise errors.ParseError(f'Ошибка парсинга {self.__class__.__name__} для {ranges}: {ex}')

    def __eq__(self, other):
        for _range in self.ranges:
            if other in _range:
                return True

        return False

    def serialize(self) -> str:
        ranges = [f'{r.start}-{r.stop}' for r in self.ranges]
        return f'{self.VALUE}{",".join(ranges)})'


class _ListValue:
    VALUE = '#list('

    def __init__(self, items: str):
        try:
            self.items = []
            for i in items.split(','):
                i = i.strip()
                if i.isdigit():
                    i = int(i)
                self.items.append(i)
        except BaseException as ex:
            raise errors.ParseError(f'Ошибка парсинга {self.__class__.__name__} для {items}: {ex}')

    def serialize(self) -> str:
        return f'{self.VALUE}{",".join(str(i) for i in self.items)})'

    def __eq__(self, other):
        return other in self.items


class BrokerTrigger(Trigger):
    TYPE = 'broker'

    def __init__(self, *args, **kwargs):
        super(BrokerTrigger, self).__init__(*args, **kwargs)
        self._topic = mqtt.topic_of(kwargs[const.TOPIC])
        _value = kwargs[const.VALUE]
        if isinstance(_value, str) and _value == _AnyValue.VALUE:
            self._value = _AnyValue()
        elif isinstance(_value, str) and _value.endswith(')') and _value.startswith(_RangeValue.VALUE):
            self._value = _RangeValue(_value[len(_RangeValue.VALUE):-1])
        elif isinstance(_value, str) and _value.endswith(')') and _value.startswith(_ListValue.VALUE):
            self._value = _ListValue(_value[len(_ListValue.VALUE):-1])
        else:
            self._value = _value

    def serialize(self) -> dict:
        res = super(BrokerTrigger, self).serialize()
        res[const.TOPIC] = str(self._topic)
        if isinstance(self._value, (_AnyValue, _RangeValue, _ListValue)):
            res[const.VALUE] = self._value.serialize()
        else:
            res[const.VALUE] = self._value

        return res

    def __str__(self):
        return f'{self._topic} as {self._value}'

    def check(self, topic: str, variable: mqtt.spread.variable.Variable):
        if str(topic) == str(self._topic):
            if variable.value == self._value:
                return True

        return False

    @property
    def topic(self) -> mqtt.spread.topic.SpreadTopic: return self._topic

    @property
    def value(self): return self._value


class DateTrigger(Trigger):
    def __init__(self, *args, **kwargs):
        super(DateTrigger, self).__init__(*args, **kwargs)
        self.time = datetime.time.fromisoformat(kwargs[const.TIME])

    @property
    def is_ready(self):
        return False

    @property
    def has_next(self):
        return False

    @property
    def time_left(self):
        return 1

    def serialize(self) -> dict:
        res = super(DateTrigger, self).serialize()
        res[const.TIME] = str(self.time)
        return res

    def __str__(self):
        return f'<{self.id}>{self.days if isinstance(self, WeekDays) else self.TYPE} at {self.time}'

    def __eq__(self, other):
        return super(DateTrigger, self).__eq__(other) and self.time == other.time

    def __cmp__(self, other):
        return self.time_left - other.time_left

    def __lt__(self, other):
        return self.time_left < other.time_left

    def __hash__(self):
        return self.id


class WeekDays(DateTrigger):
    TYPE = 'weekdays'

    def __init__(self, *args, **kwargs):
        super(WeekDays, self).__init__(*args, **kwargs)
        days = kwargs[const.DAYS]
        self.days = []
        if len(days) == 0:
            raise errors.HandlingError(f'{const.DAYS} expected 1..7 items')
        for day in days:
            if day in WEEK:
                self.days.append(day)
            else:
                raise errors.HandlingError(f'{day} not include in standard values {WEEK}')

    def serialize(self) -> dict:
        res = super(WeekDays, self).serialize()
        res[const.DAYS] = self.days
        return res

    @property
    def has_next(self):
        return True

    @property
    def time_left(self):
        now = datetime.datetime.now()
        days = []
        for day in self.days:
            index = WEEK.index(day)
            if index == now.weekday() and self.time > now.time():
                delta = 0
            elif index > now.weekday() or (index == now.weekday() and self.time > now.time()):
                delta = index - now.weekday()
            else:
                delta = 7 - now.weekday() + index
            _date = (now + datetime.timedelta(days=delta)).date()
            t = datetime.datetime.fromisoformat(str(_date) + ' ' + str(self.time))
            days.append(t)

        days.sort()

        target_time = days[0].timestamp()
        return int(target_time - now.timestamp())


class Holiday(DateTrigger):
    TYPE = 'holiday'
    _days = None

    @staticmethod
    def days():
        if Holiday._days is None:
            Holiday._days = generate_holidays()
        return Holiday._days

    @property
    def has_next(self):
        return True

    @property
    def time_left(self):
        now = datetime.datetime.now()

        days = self.days().copy()

        while len(days) > 0:
            if days[0] < now.date() or (days[0] == now.date() and now.time() > self.time):
                days.pop(0)
            else:
                break

        if len(days) > 0:
            _date = days[0]
        else:
            raise errors.HandlingError('calendar is empty!')

        target_time = datetime.datetime.fromisoformat(str(_date) + ' ' + str(self.time))
        return int(target_time.timestamp() - now.timestamp())


class WorkDays(DateTrigger):
    TYPE = 'workday'

    @property
    def has_next(self):
        return True

    @property
    def time_left(self):
        now = datetime.datetime.now()

        _date = now.date()

        if _date not in Holiday.days():
            if self.time <= now.time():
                _date += datetime.timedelta(days=1)

        while _date in Holiday.days():
            _date += datetime.timedelta(days=1)

        target_time = datetime.datetime.fromisoformat(str(_date) + ' ' + str(self.time))
        return int((target_time - now).total_seconds())


def date_trigger_of_data(_type: str, data: dict) -> DateTrigger:
    for cl in DateTrigger.__subclasses__():
        if issubclass(cl, DateTrigger):
            if cl.TYPE == _type:
                return cl(**data)


def of(data: dict) -> Trigger:
    def separator(subclass): return issubclass(subclass, Trigger) and subclass.TYPE == data[const.TYPE]
    trigger_class = utils.get_subclass(Trigger, separator=separator)
    return trigger_class(**data)
